//#include "includes.h"
#include <stdint.h>
#include "gpio.h"
#include "sf_uart.h"

void (*sf_uart0_appRxEventHandler)(uint16_t rx);
void (*sf_uart1_appRxEventHandler)(uint16_t rx);
void (*sf_uart2_appRxEventHandler)(uint16_t rx);
void (*sf_uart3_appRxEventHandler)(uint16_t rx);


sf_uart_config sf_uart0_config =
{
    .rxPin = SF_RX0_PIN_NUMBER,
    .txPin = SF_TX0_PIN_NUMBER,
    .baudrate = 9600,
    .delay =
        {
            .rx_centering = 0,
            .rx_intrabit = 0,
            .rx_stopbit = 0,
            .tx = 0,
        }
};

sf_uart_config sf_uart1_config =
{
    .rxPin = SF_RX1_PIN_NUMBER,
    .txPin = SF_TX1_PIN_NUMBER,
    .baudrate = 9600,
    .delay =
        {
            .rx_centering = 0,
            .rx_intrabit = 0,
            .rx_stopbit = 0,
            .tx = 0,
        }
};

sf_uart_config sf_uart2_config =
{
    .rxPin = SF_RX2_PIN_NUMBER,
    .txPin = SF_TX2_PIN_NUMBER,
    .baudrate = 9600,
    .delay =
        {
            .rx_centering = 0,
            .rx_intrabit = 0,
            .rx_stopbit = 0,
            .tx = 0,
        }
};

sf_uart_config sf_uart3_config =
{
    .rxPin = SF_RX3_PIN_NUMBER,
    .txPin = SF_TX3_PIN_NUMBER,
    .baudrate = 9600,
    .delay =
        {
            .rx_centering = 0,
            .rx_intrabit = 0,
            .rx_stopbit = 0,
            .tx = 0,
        }
};

static sf_uart_config *configs[] =
{
    &sf_uart0_config, &sf_uart1_config,
    &sf_uart2_config, &sf_uart3_config,
};

void sf_uart0_drvRxEventHandler()
{
    uint8_t rx = 0;
    rx = sf_uart_drv_recv(configs[APP_SF_UART0]);
    sf_uart0_appRxEventHandler(rx);
}

void sf_uart1_drvRxEventHandler()
{
    uint8_t rx = 0;
    rx = sf_uart_drv_recv(configs[APP_SF_UART1]);
    sf_uart0_appRxEventHandler(rx);
}

void sf_uart2_drvRxEventHandler()
{
    uint8_t rx = 0;
    rx = sf_uart_drv_recv(configs[APP_SF_UART2]);
    sf_uart0_appRxEventHandler(rx);
}

void sf_uart3_drvRxEventHandler()
{
    uint8_t rx = 0;
    rx = sf_uart_drv_recv(configs[APP_SF_UART3]);
    sf_uart0_appRxEventHandler(rx);
}

static sf_uart_rxEventHandler sf_uart_drvRxEventHandlers[] =
{
    &sf_uart0_drvRxEventHandler, &sf_uart1_drvRxEventHandler,
    &sf_uart2_drvRxEventHandler, &sf_uart3_drvRxEventHandler,
};

static EventHandler *sf_uart_appRxEventHandlers[] =
{
    &sf_uart0_appRxEventHandler, &sf_uart1_appRxEventHandler,
    &sf_uart2_appRxEventHandler, &sf_uart3_appRxEventHandler,
};

void delay_loop(uint32_t cycles)
{
  volatile uint32_t cycle = cycles;
  for(; cycle>0;cycle--)
  {
      __NOP();
  }
}


uint16_t subtract_cap(uint16_t num, uint16_t sub) 
{
  volatile uint16_t diff = 0;
  if (num > sub)
  {
    diff = num - sub; //- 6;
    return diff;
  }
  else
    return 1;
}


pin_status sf_uart_drv_init(sf_uart_config *config, sf_uart_rxEventHandler event_handler)
{
    // Precalculate the various delays, in number of 6-cycle delays
    uint16_t bit_delay = (F_XTAL / config->baudrate)/6;

    // 12 (gcc 4.8.2) or 13 (gcc 4.3.2) cycles from start bit to first bit,
    // 15 (gcc 4.8.2) or 16 (gcc 4.3.2) cycles between bits,
    // 12 (gcc 4.8.2) or 14 (gcc 4.3.2) cycles from last bit to stop bit
    // These are all close enough to just use 15 cycles, since the inter-bit
    // timings are the most critical (deviations stack 8 times)
    config->delay.tx = subtract_cap(bit_delay, 15/6);
    // Timings counted from gcc 4.8.2 output. This works up to 115200 on
    // 16Mhz and 57600 on 8Mhz.
    //
    // When the start bit occurs, there are 3 or 4 cycles before the
    // interrupt flag is set, 4 cycles before the PC is set to the right
    // interrupt vector address and the old PC is pushed on the stack,
    // and then 75 cycles of instructions (including the RJMP in the
    // ISR vector table) until the first delay. After the delay, there
    // are 17 more cycles until the pin value is read (excluding the
    // delay in the loop).
    // We want to have a total delay of 1.5 bit time. Inside the loop,
    // we already wait for 1 bit time - 23 cycles, so here we wait for
    // 0.5 bit time - (71 + 18 - 22) cycles.
    config->delay.rx_centering = subtract_cap(bit_delay / 2, (4 + 4 + 75 + 17 - 23)/6) - 7;//corrections for nRF;

    // There are 23 cycles in each loop iteration (excluding the delay)
    config->delay.rx_intrabit = subtract_cap(bit_delay, 23/6) - 6;//corrections for nRF;

    // There are 37 cycles from the last bit read to the start of
    // stopbit delay and 11 cycles from the delay until the interrupt
    // mask is enabled again (which _must_ happen during the stopbit).
    // This delay aims at 3/4 of a bit time, meaning the end of the
    // delay will be at 1/4th of the stopbit. This allows some extra
    // time for ISR cleanup, which makes 115200 baud at 16Mhz work more
    // reliably
    config->delay.rx_stopbit = subtract_cap(bit_delay*3/4, (37 + 11)/6) - 15;//corrections for nRF;
    
    pin_status p_status = PIN_VALID;
    p_status = gpiote_pinMode(config->rxPin, INPUT, GPIOTE_CONFIG_POLARITY_HiToLo, event_handler);
    while(p_status != PIN_VALID);
    p_status = gpio_pinMode(config->txPin, OUTPUT);
    while(p_status != PIN_VALID);
    gpio_write(config->txPin, HIGH);
}


bool sf_uart_drv_write(sf_uart_config *config, uint8_t x)
{
    if (config->delay.tx == 0) 
    {
        return 0;
    }
  
    //Write the start bit
    gpio_write(config->txPin, LOW);
    delay_loop(config->delay.tx);

    // Write each of the 8 bits
    for (uint8_t i = 8; i > 0; --i)
    {
      if (x & 1) // choose bit
        gpio_write(config->txPin, HIGH);
      else
        gpio_write(config->txPin, LOW);

      delay_loop(config->delay.tx);
      x >>= 1;
    }

    //restore pin to natural state
    gpio_write(config->txPin, HIGH);
    delay_loop(config->delay.tx);

    return 1;
}

uint8_t sf_uart_drv_recv(sf_uart_config *config)
{
  uint8_t d = 0;

  // If RX line is high, then we don't see any start bit
  // so interrupt is probably not for us
  if (!gpio_read(config->rxPin))
  {
    // Disable further interrupts during reception, this prevents
    // triggering another interrupt directly after we return, which can
    // cause problems at higher baudrates.
    gpiote_eventDisable(config->rxPin);
        
    // Wait approximately 1/2 of a bit width to "center" the sample
    delay_loop(config->delay.rx_centering);

    // Read each of the 8 bits
    for(uint8_t i=8; i > 0; --i)
    {
      delay_loop(config->delay.rx_intrabit);       
      d >>= 1;
      if (gpio_read(config->rxPin))
        d |= 0x80;
    }
  
    // skip the stop bit
    delay_loop(config->delay.rx_stopbit);

    // Re-enable interrupts when we're sure to be inside the stop bit
    gpiote_eventEnable(config->rxPin);
    return d;
  }
}

void sf_uart_init(int uartNumb, sf_uart_baud_rate baud, EventHandler rx_handler)
{
    configs[uartNumb]->baudrate = baud;
    sf_uart_drv_init(configs[uartNumb], sf_uart_drvRxEventHandlers[uartNumb]);
    *sf_uart_appRxEventHandlers[uartNumb] = rx_handler;
}

void sf_uart_write(int uartNumb,  uint8_t x)
{
   sf_uart_drv_write(configs[uartNumb], x);
}

void sf_uart_writeStr(int uartNumb, char* str)
{
    uint16_t i = 0;
    while (str[i] != NULL)
    {
        sf_uart_write(uartNumb, str[i]);
        i++;
    }
}

void sf_uart_writeBuf(int uartNumb, uint8_t *data, int size)
{
    for(int i = 0; i < size; i++ )
    {
        sf_uart_write(uartNumb, *data++);
    }
}

void sf_uart_rxEnable(int uartNumb)
{
    gpiote_eventEnable(configs[uartNumb]->rxPin);
}

void sf_uart_rxDisable(int uartNumb)
{
    gpiote_eventDisable(configs[uartNumb]->rxPin);
}
